""" You can kind of see this as the scope of `just` when you 'import just'
The following functions become available:
just.__project__
just.__version__
just.run
just.print_version
"""

import os
import errno
import shutil
from just.path_ import (
    ls,
    rename,
    exists,
    make_path,
    get_just_path,
    glob,
    remove,
    mkdir,
    most_recent,
)
from just.read_write import *
from just.requests_ import request, request_tree, get, get_tree, post, post_tree, save_session
from just.dir import mkdir
from just.log import log
from just.jpath import json_extract, jpath
from just.pattern import Pattern

# In [19]: with open("fuck.xz", "wb") as f: f.write(lzma.compress(html.encode()))

# In [20]: with lzma.open("fuck.xz") as f: hh=f.read()

__project__ = "just"
__version__ = "0.8.159"



def read(fname):
    with open(fname) as f:
        return [x for x in f.read().split("\n") if x]


def iread(fname):
    with open(fname) as f:
        for line in f:
            if line:
                yield line.strip()


def write(obj, fname):
    with open(fname, "w") as f:
        f.write("\n".join(obj))



import warnings
from just.dt import parse_dt
from collections import defaultdict
import orjson
import json

parse_knowledge = defaultdict(lambda: (0, 0))


def default(obj):
    if isinstance(obj, set):
        return list(obj)
    tp = str(type(obj))
    if ".float" in tp:
        return float(obj)
    if ".int" in tp:
        return int(obj)
    raise TypeError


def handle_str_value(path, v):
    hits, total = parse_knowledge[path]
    if total == 5:
        if hits == 5:
            return parse_dt(v)
        return v
    try:
        v = parse_dt(v)
        hits += 1
    except ValueError:
        pass
    parse_knowledge[path] = (hits, total + 1)
    return v


def try_dt(col):
    try:
        return [parse_dt(x) if isinstance(x, str) else x for x in col]
    except ValueError:
        return col


def parse_datetimes(obj, path: str = ""):
    if isinstance(obj, str):
        return handle_str_value(path, obj)
    if isinstance(obj, list):
        return [parse_datetimes(x, "." + path) for x in obj]
    if isinstance(obj, dict):
        return {k: parse_datetimes(v, path + k) for k, v in obj.items()}
    return obj


def read(fn, warn=False, parse_dt=False):
    if not isinstance(fn, str):
        return json.load(fn)
    if fn.endswith(".jsonl"):
        if warn:
            warnings.warn("Reading streaming format at once.")
        data = list(iread(fn))
        if parse_dt:
            data = parse_datetimes(data)
        return data
    with open(fn) as f:
        data = json.load(f)
        if parse_dt:
            data = parse_datetimes(data)
        return data


def append(obj, fn):
    if not isinstance(fn, str):
        raise TypeError("Cannot append to compression")
    with open(fn, "a+") as f:
        f.write(orjson.dumps(obj, default=default).decode() + "\n")


def write(obj, fn, option=orjson.OPT_INDENT_2):
    if not isinstance(obj, bytes):
        obj = orjson.dumps(obj, option=option, default=default)
    if not isinstance(fn, str):
        fn.write(obj)
    else:
        with open(fn, "w") as f:
            f.write(obj.decode())


def iread(fn):
    if not isinstance(fn, str):
        raise TypeError("Cannot iteratively read compressed file now")
    with open(fn) as f:
        for i, line in enumerate(f):
            try:
                data = orjson.loads(line)
                if parse_dt:
                    data = parse_datetimes(data)
                yield data
            except Exception as e:
                msg = "JSON-L parsing error in line number {} in the jsonl file".format(i)
                raise Exception(msg, line)


def iwrite(obj, fn):
    if not isinstance(fn, str):
        raise TypeError("Cannot iteratively write compressed")
    with open(fn, "w") as f:
        for chunk in obj:
            f.write(orjson.dumps(chunk, default=default).decode() + "\n")



""" Our tests are defined in here """
import os
import sys
from operator import eq
import just
import pytest


TEST_FNAME = "testobj"


def get_result(m, extension, inp):
    fname = TEST_FNAME + "." + extension
    try:
        m.write(inp, fname)
        read_result = m.read(fname)
    finally:
        just.remove(fname)
    return read_result


@pytest.mark.parametrize(
    "m, extension, inp, expected, compare",
    [
        (just, "txt", "{}", "{}", eq),
        (just.txt, "txt", "{}", "{}", eq),
        (just.txt, "txt", "None", "None", eq),
        (just.txt, "txt", "", "", eq),
        (just.newl, "newl", ["1", "2"], ["1", "2"], eq),
        (just.json, "json", {}, {}, eq),
        (just.json, "json", None, None, eq),
        (just.json, "json", "", "", eq),
        (just.yaml, "yaml", {}, {}, eq),
        (just.yaml, "yaml", None, None, eq),
        (just.yaml, "yaml", "", "", eq),
        (just.pickle, "pkl", {}, {}, eq),
        (just.pickle, "pkl", None, None, eq),
        (just.pickle, "pkl", "", "", eq),
    ],
)
def test_compare(m, extension, inp, expected, compare):
    assert compare(get_result(m, extension, inp), expected)


def test_multi_read():
    obj = ["a", "b"]
    fnames = ["a.txt", "b.txt"]
    just.multi_write(obj, fnames)
    try:
        for name, data in just.multi_read("*.txt"):
            assert fnames.index(name.split("/")[-1]) == obj.index(data)
    finally:
        for fname in fnames:
            os.remove(fname)


def test_newl_iread():
    fname = "testobj.newl"
    obj = ["1", "2"]
    just.write(obj, "testobj.newl")
    try:
        assert [x for x in just.iread(fname)] == [x for x in obj]
    finally:
        os.remove(fname)


def test_txt_iread():
    fname = "testobj.txt"
    obj = "1\n2\n3\n4\n5"
    just.write(obj, "testobj.txt")
    try:
        assert [x for x in just.iread(fname)] == [x for x in obj.split("\n")]
    finally:
        os.remove(fname)


def test_find_just_path():
    try:
        base = os.path.dirname(os.path.abspath(__file__))
        just_file = os.path.join(base, ".just")
        with open(just_file, "w") as f:
            f.write("OK")
        assert isinstance(just.path_.find_just_path(), type("1"))
    finally:
        os.remove(just_file)


def test_txt_append():
    fname = "testobj.txt"
    obj = "bla"
    just.append(obj, "testobj.txt")
    try:
        assert [x for x in just.iread(fname)] == [obj]
        just.append(obj, "testobj.txt")
        assert [x for x in just.iread(fname)] == [obj, obj]
    finally:
        os.remove(fname)


def test_unsuccesful_read():
    assert just.read("A" * 100, 42) == 42


def test_unsuccesful_remove():
    assert just.remove("A" * 100, 42) == 42


def test_ls():
    assert just.ls(".")



import just
import requests_mock
import pytest

NOT_EXISTING_URL = "mock://stuasdfasfasdfasf.com"


@pytest.mark.parametrize(
    "method, output, content_type",
    [
        ("GET", "mock_text", "text/html"),
        ("GET", {"mock": "text"}, "application/json"),
        ("POST", "mock_text", "text/html"),
        ("POST", {"mock": "text"}, "application/json"),
    ],
)
def test_requests(method, output, content_type):
    with requests_mock.mock() as m:
        headers = {"Content-Type": content_type}
        just_method = just.get if method == "GET" else just.post
        m_method = m.get if method == "GET" else m.post
        if content_type == "application/json":
            m_method(NOT_EXISTING_URL, json=output, headers=headers)
            assert just_method(NOT_EXISTING_URL) == output
        else:
            m_method(NOT_EXISTING_URL, text=output, headers=headers)
            assert just_method(NOT_EXISTING_URL) == output


def test_retry():
    assert not just.get(NOT_EXISTING_URL, max_retries=2, delay_base=0.01)



from ciso8601 import parse_datetime as parse_dt
from datetime import datetime, timedelta, timezone


def utcnow(**kwargs):
    if kwargs:
        return datetime.now(timezone.utc) + timedelta(**kwargs)
    return datetime.now(timezone.utc)



import inspect
import os
import glob2
import errno
import functools

__cached_just_path = None


def glob(path, sort_reverse=False):
    path = make_path(path)
    return sorted(glob2.glob(path), reverse=sort_reverse)


def get_just_env_path():
    return os.environ.get("JUST_PATH")


def find_just_path(base=None, max_depth=5):
    if base is None:
        base = os.path.dirname(os.path.abspath(inspect.stack()[1][1]))
    for depth in range(max_depth):
        prefix = "../" * depth
        just_file = os.path.join(base, prefix, ".just")
        if os.path.isfile(just_file):
            print("just_file", os.path.abspath(os.path.dirname(just_file)))
            return os.path.abspath(os.path.dirname(just_file))
    return None


def get_likely_path():
    # try:
    # main_path = os.path.abspath(sys.modules['__main__'].__file__)
    # except AttributeError:
    main_path = os.path.realpath("__file__")
    return os.path.dirname(main_path)


def get_just_path():
    global __cached_just_path
    if __cached_just_path is not None:
        return __cached_just_path
    just_path = get_just_env_path()
    just_path = just_path if just_path is not None else find_just_path()
    just_path = just_path if just_path is not None else find_just_path(".")
    just_path = just_path if just_path is not None else get_likely_path()
    __cached_just_path = just_path
    return just_path


@functools.lru_cache(maxsize=1_000_000)
def make_path(filename):
    just_path = get_just_path()
    if not isinstance(filename, (str, bytes)):
        filename = filename.name.encode("utf8").decode()
    filename = filename.replace("file://", "")
    path = os.path.join(just_path, os.path.expanduser(filename))
    if path.endswith("."):
        path = path[:-1]
    return path


def exists(fname):
    return os.path.isfile(make_path(fname))


def rename(src, dest, no_exist=None):
    src = make_path(src)
    dest = make_path(dest)
    if not dest.endswith("/"):
        *dest_folder_parts, filename = dest.split("/")
        dest_folder = "/".join(dest_folder_parts)
        if not os.path.exists(dest_folder) and "." in filename:
            os.makedirs(dest_folder)
    if not os.path.isfile(src) and no_exist is not None:
        return False
    os.rename(src, dest)
    return True


def _as_glob(dir_name, recursive):
    dir_name = make_path(dir_name)
    if not "*" in dir_name:
        if dir_name.endswith("/"):
            dir_name += "*"
        else:
            dir_name += "/*"
        if recursive:
            dir_name += "*"
    return dir_name


def ls(dir_name, recursive=False, no_dirs=False):
    dir_name = _as_glob(dir_name, recursive)
    if no_dirs:
        return [x for x in glob(dir_name) if not os.path.isdir(x)]
    else:
        return glob(dir_name)


def mkdir(path, mode=0o777):
    path = make_path(path)
    try:
        os.makedirs(path, mode)
    # Python >2.5
    except OSError as exc:  # pragma: no cover
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def remove(file_path, no_exist=False, allow_recursive=False):
    if isinstance(file_path, (tuple, list)):
        file_path = os.path.join(*file_path)
    if "*" in file_path:
        if not allow_recursive:
            raise IOError("Cannot remove wildcard unless allow_recursive=True")
        paths = glob(file_path)
        for fn in sorted(paths, key=lambda x: -len(x)):
            os.remove(fn)
        return bool(paths)
    file_path = make_path(file_path)
    if os.path.isfile(file_path):
        os.remove(file_path)
        return True
    if os.path.isdir(file_path):
        if allow_recursive:
            shutil.rmtree(file_path)
            return True
        else:
            raise IOError("Cannot remove directory unless allow_recursive=True")
    # if there is a default value, return that if no file/dir found when attempting to remove
    if no_exist is not None:
        return no_exist
    raise IOError("File '{}' does not exist.".format(file_path))


def most_recent(file_path):
    return max(glob(file_path), key=os.path.getctime)



def read(fname):
    if not isinstance(fname, str):
        return fname.read()
    with open(fname, "rb") as f:
        return f.read()


def iread(fname):
    if not isinstance(fname, str):
        raise TypeError("Cannot iteratively read compressed file at this point.")
    with open(fname, "rb") as f:
        for char in f:
            yield char


def write(obj, fname):
    if not isinstance(fname, str):
        fname.write(obj)
    else:
        with open(fname, "wb") as f:
            f.write(obj)



import re
from ciso8601 import parse_datetime as isoparse
from datetime import datetime
from just import iread


def parse_specials(special_str):
    if not special_str:
        return "", "str"
    if "@" not in special_str:
        return special_str, ""
    return special_str.split("@")


def iso(x):
    return datetime.fromisoformat(x.replace(",", ".").replace("Z", "+00:00"))


def convert_str(format_str):
    r = ""
    on = False
    specials = []
    types = []
    tmp = ""
    lenm = len(format_str)
    last = -1
    captured_names = []
    for i, x in enumerate(format_str):
        if x == "}":
            on = False
            # fw = False
            name, cast = parse_specials(tmp)
            capture = bool(name)
            tp = eval(cast) if cast else str
            # if fw:
            #     pat = ".{4}"
            if i + 1 == lenm:
                pat = ".+"
            else:
                pat = f"[^{format_str[i+1]}]+"
            if capture:
                pat = f"({pat})"
                captured_names.append(name)
                types.append(tp)
            r += pat
            tmp = ""
            last = len(r)
        elif x == "{":
            on = True
        else:
            if not on:
                r += re.escape(x)
            else:
                tmp += x
    return r, types, captured_names


NUMERIC = {int, float}


class Pattern:
    def __init__(self, format_str):
        self.format_str = format_str
        self.pattern, self.types, self.names = convert_str(format_str)
        if len(self.types) == 1:
            self.type = self.types[0]
            self.find = self.finder_one
        else:
            self.find = self.finder_multi
        self.r = re.compile(self.pattern)
        self.num_captures_vars = len([x for x in self.names if x])

    def finder_one(self, line):
        res = self.r.search(line)
        if not res:
            return None
        return self.type(res.groups()[0].strip())

    def finder_multi(self, line):
        res = self.r.search(line)
        if not res:
            return [None] * self.num_captures_vars
        return [tp(x.strip()) for name, tp, x in zip(self.names, self.types, res.groups()) if name]

    def find_dict(self, line):
        res = self.finder_multi(line)
        if res[0] is None:
            return None
        return {k: v for k, v in zip(self.names, res) if isinstance(k, int) or not k.startswith("_") and k}

    def stream(self, fname):
        return [self.find_dict(x) for x in iread(fname)]



import time
import sys
import inspect
import just

START = "_".join(time.asctime().replace(":", "_").split())
NAME = sys.argv[0].rstrip(".py")
LOG_BASE = "logs"
LOG_FILE = "{}/{}_{}.jsonl".format(LOG_BASE, NAME, START)
LOG_LINK = "{}/{}.jsonl".format(LOG_BASE, NAME)


def get_file_scope():
    frame = inspect.stack()[2]
    file_name, function_name = frame[1], frame[3]
    return file_name, function_name


def log(obj, *tags):
    file_name, function_name = get_file_scope()
    data = {"tags": tags, "object": obj, "file": file_name, "function": function_name,
            "time": time.time()}
    just.append(data, LOG_FILE)



def read(fname):
    if not isinstance(fname, str):
        return fname.read().decode()
    with open(fname) as f:
        return f.read()


def iread(fname):
    if not isinstance(fname, str):
        raise TypeError("Cannot iteratively read compressed file at this point.")
    with open(fname) as f:
        for line in f:
            yield line.rstrip("\n")


def write(obj, fname):
    if not isinstance(fname, str):
        fname.write(obj.encode())
    else:
        with open(fname, "w") as f:
            f.write(obj)


def append(obj, fname):
    if not isinstance(fname, str):
        raise TypeError("Cannot append to compression")
    with open(fname, "a+") as f:
        f.write(obj + "\n")



try:
    import zstandard as zstd
except ImportError:
    import zstd


class ZstdFile(object):
    def __init__(self, fname, *ignore, **kw_ignore):
        self.fname = fname

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if value:
            raise value
        return True

    def read(self, **kwargs):
        with open(self.fname, "rb") as f:
            return zstd.ZstdDecompressor().decompress(f.read())

    def write(self, obj, **kwargs):
        with open(self.fname, "wb") as f:
            if isinstance(obj, str):
                obj = bytes(obj, encoding="utf8")
            f.write(zstd.ZstdCompressor().compress(obj))



from just.read_write import read


def json_extract(dc, expr):
    from jsonpath_rw import parse

    res = parse(expr).find(dc)
    if len(res) == 1:
        res = res[0].value
    else:
        res = [x.value for x in res]
    return res


def jpath(fname_or_dc, jsonpath_expression):
    if isinstance(fname_or_dc, str):
        fname_or_dc = read(fname_or_dc)
    return json_extract(fname_or_dc, jsonpath_expression)
